#!/usr/bin/env python2
# -*- coding: utf-8 -*-

# Contest Management System - http://cms-dev.github.io/
# Copyright © 2011-2013 Luca Wehrstedt <luca.wehrstedt@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import io
import json
import os
import pkg_resources
import sys


class Config(object):
    """An object holding the current configuration.

    """
    def __init__(self):
        """Fill this object with the default values for each key.

        """
        # Connection.
        self.bind_address = ''
        self.http_port = 8890
        self.https_port = None
        self.https_certfile = None
        self.https_keyfile = None
        self.timeout = 600  # 10 minutes (in seconds)

        # Authentication.
        self.realm_name = 'Scoreboard'
        self.username = 'usern4me'
        self.password = 'passw0rd'

        # Buffers
        self.buffer_size = 100  # Needs to be strictly positive.

        # File system.
        self.installed = sys.argv[0].startswith("/usr/") and \
            sys.argv[0] != '/usr/bin/ipython' and \
            sys.argv[0] != '/usr/bin/python2' and \
            sys.argv[0] != '/usr/bin/python'

        self.web_dir = pkg_resources.resource_filename("cmsranking", "static")
        if self.installed:
            self.log_dir = os.path.join("/", "var", "local", "log",
                                        "cms", "ranking")
            self.lib_dir = os.path.join("/", "var", "local", "lib",
                                        "cms", "ranking")
            paths = [os.path.join("/", "usr", "local", "etc",
                                  "cms.ranking.conf"),
                     os.path.join("/", "etc", "cms.ranking.conf")]
        else:
            self.log_dir = os.path.join("log", "ranking")
            self.lib_dir = os.path.join("lib", "ranking")
            paths = [os.path.join(".", "config", "cms.ranking.conf"),
                     os.path.join("/", "usr", "local", "etc",
                                  "cms.ranking.conf"),
                     os.path.join("/", "etc", "cms.ranking.conf")]

        try:
            os.makedirs(self.lib_dir)
        except OSError:
            pass  # We assume the directory already exists...

        try:
            os.makedirs(self.web_dir)
        except OSError:
            pass  # We assume the directory already exists...

        try:
            os.makedirs(self.log_dir)
        except OSError:
            pass  # We assume the directory already exists...

        self._load(paths)

    def get(self, key):
        """Get the config value for the given key.

        """
        return getattr(self, key)

    def _load(self, paths):
        """Try to load the config files one at a time, until one loads
        correctly.

        """
        for conf_file in paths:
            try:
                self._load_unique(conf_file)
            except IOError:
                # We cannot access the file, we skip it.
                pass
            except ValueError as exc:
                print("Unable to load JSON configuration file %s, probably "
                      "because of a JSON decoding error.\n%r" % (conf_file,
                                                                 exc))
            else:
                print("Using configuration file %s." % conf_file)
                return
        print("Warning: no configuration file found.")

    def _load_unique(self, path):
        """Populate the Config class with everything that sits inside
        the JSON file path (usually something like /etc/cms.conf). The
        only pieces of data treated differently are the elements of
        core_services and other_services that are sent to async
        config.

        path (string): the path of the JSON config file.

        """
        # Load config file
        with io.open(path, 'rb') as fobj:
            data = json.load(fobj)

            # Put everything.
            for key, value in data.iteritems():
                setattr(self, key, value)


# Create an instance of the Config class.
config = Config()



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

# Contest Management System - http://cms-dev.github.io/
# Copyright © 2011-2013 Luca Wehrstedt <luca.wehrstedt@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import heapq
import logging

from cmsranking.Submission import store as submission_store
from cmsranking.Subchange import store as subchange_store
from cmsranking.Task import store as task_store


logger = logging.getLogger(__name__)


class NumberSet(object):
    """A fast data structure on numbers.

    It supports:
    - inserting a value
    - removing a value
    - querying the maximum value

    It can hold the same value multiple times.

    This data structure could be implemented with a binary tree, but
    at the moment we're actually using a standard python list.

    """
    def __init__(self):
        self._impl = list()

    def insert(self, val):
        self._impl.append(val)

    def remove(self, val):
        self._impl.remove(val)

    def query(self):
        return max(self._impl + [0.0])

    def clear(self):
        del self._impl[:]


class Score(object):
    """The score of a user for a task.

    It computes the current score (and its history) for this
    user/task.  It gets notified in case a submission is created,
    updated and deleted.

    """
    # We assume that the submissions will all have different times,
    # since cms enforces a minimum delay between two submissions of
    # the same user for the same task.
    # On the other hand, different subchanges may have the same time
    # but cms assures that the order in which the subchanges have to
    # be processed is the ascending order of their keys (actually,
    # this is enforced only for subchanges with the same time).
    def __init__(self, score_mode="max_tokened_last"):
        # The submissions in their current status.
        self._submissions = dict()

        # The list of changes of the submissions.
        self._changes = list()

        # The set of the scores of the currently released submissions.
        self._released = NumberSet()

        # The last submitted submission (with at least one subchange).
        self._last = None

        # The history of score changes (the actual "output" of this
        # object).
        self._history = list()

        self._score_mode = score_mode

    def append_change(self, change):
        # Remove from released submission (if needed), apply changes,
        # add back to released submissions (if needed) and check if
        # it's the last. Compute the new score and, if it changed,
        # append it to the history.
        s_id = change.submission
        if self._submissions[s_id].token:
            self._released.remove(self._submissions[s_id].score)
        if change.score is not None:
            self._submissions[s_id].score = change.score
        if change.token is not None:
            self._submissions[s_id].token = change.token
        if change.extra is not None:
            self._submissions[s_id].extra = change.extra
        if self._submissions[s_id].token:
            self._released.insert(self._submissions[s_id].score)
        if change.score is not None and \
                (self._last is None or
                 self._submissions[s_id].time > self._last.time):
            self._last = self._submissions[s_id]

        if self._score_mode == "max":
            score = max([0.0] +
                        [submission.score
                         for submission in self._submissions.values()])
        else:
            score = max(self._released.query(),
                        self._last.score if self._last is not None else 0.0)

        if score != self.get_score():
            self._history.append((change.time, score))

    def get_score(self):
        return self._history[-1][1] if len(self._history) > 0 else 0.0

    def reset_history(self):
        # Delete everything except the submissions and the subchanges.
        self._last = None
        self._released.clear()
        del self._history[:]

        # Reset the submissions at their default value.
        for sub in self._submissions.itervalues():
            sub.score = 0.0
            sub.token = False
            sub.extra = list()

        # Append each change, one at a time.
        for change in self._changes:
            self.append_change(change)

    def create_subchange(self, key, subchange):
        # Insert the subchange at the right position inside the
        # (sorted) list and call the appropriate method (append_change
        # or reset_history)
        if len(self._changes) == 0 or \
                subchange.time > self._changes[-1].time or \
                (subchange.time == self._changes[-1].time and
                 subchange.key > self._changes[-1].key):
            self._changes.append(subchange)
            self.append_change(subchange)
        else:
            for idx, val in enumerate(self._changes):
                if subchange.time < val.time or \
                   (subchange.time == val.time and subchange.key < val.key):
                    self._changes.insert(idx, subchange)
                    break
            self.reset_history()
            logger.info("Reset history for user '%s' and task '%s' after "
                        "creating subchange '%s' for submission '%s'",
                        self._submissions[subchange.submission].user,
                        self._submissions[subchange.submission].task,
                        key, subchange.submission)

    def update_subchange(self, key, subchange):
        # Update the subchange inside the (sorted) list and,
        # regardless of its position in that list, reset the history.
        for i in range(len(self._changes)):
            if self._changes[i].key == key:
                self._changes[i] = subchange
        self.reset_history()
        logger.info("Reset history for user '%s' and task '%s' after "
                    "creating subchange '%s' for submission '%s'",
                    self._submissions[subchange.submission].user,
                    self._submissions[subchange.submission].task,
                    key, subchange.submission)

    def delete_subchange(self, key):
        # Delete the subchange from the (sorted) list and reset the
        # history.
        self._changes = [c for c in self._changes if c.key != key]
        self.reset_history()
        logger.info("Reset history after deleting subchange '%s'", key)

    def create_submission(self, key, submission):
        # A new submission never triggers an update in the history,
        # since it doesn't have a score.
        submission.score = 0.0
        submission.token = False
        submission.extra = list()
        self._submissions[key] = submission

    def update_submission(self, key, submission):
        # An updated submission may cause an update in history because
        # it may change the "last" submission at some point in
        # history.
        self._submissions[key] = submission
        self.reset_history()

    def delete_submission(self, key):
        # A deleted submission shouldn't cause any history changes
        # (because its associated subchanges are deleted before it)
        # but we reset it just to be sure...
        if key in self._submissions:
            del self._submissions[key]
            # Delete all its subchanges.
            self._changes = [c for c in self._changes if c.submission != key]
            self.reset_history()

    def update_score_mode(self, score_mode):
        self._score_mode = score_mode


class ScoringStore(object):
    """A manager for all instances of Scoring.

    It listens to the events of submission_store and subchange_store and
    redirects them to the corresponding Score (based on their user/task).
    When asked to provide a global history of score changes it takes the
    ones of each Score and combines them toghether (using a binary heap).

    """
    # We can do an important assumption here too: since the data has
    # to be consistent we are sure that if there's at least one
    # subchange there's also at least one submission (and if there's
    # no submission there's no subchange). We can also assume that
    # when a submission is deleted its subchanges have already been
    # deleted. So we are sure that we can delete the Score after we
    # delete the last submission, but we cannot after we delete the
    # last subchange.
    def __init__(self):
        submission_store.add_create_callback(self.create_submission)
        submission_store.add_update_callback(self.update_submission)
        submission_store.add_delete_callback(self.delete_submission)
        subchange_store.add_create_callback(self.create_subchange)
        subchange_store.add_update_callback(self.update_subchange)
        subchange_store.add_delete_callback(self.delete_subchange)

        self._scores = dict()
        self._callbacks = list()

    def init_store(self):
        """Load the scores from the stores.

        This method must be called by RankingWebServer after it
        finishes loading the data from disk.

        """
        for key, value in submission_store._store.iteritems():
            self.create_submission(key, value)
        for key, value in sorted(subchange_store._store.iteritems()):
            self.create_subchange(key, value)

    def add_score_callback(self, callback):
        """Add a callback to be called when a score changes.

        Callbacks can be any kind of callable objects. They must
        accept three arguments: the user, the task and the new score.

        """
        self._callbacks.append(callback)

    def notify_callbacks(self, user, task, score):
        for call in self._callbacks:
            call(user, task, score)

    def create_submission(self, key, submission):
        if submission.user not in self._scores:
            self._scores[submission.user] = dict()
        if submission.task not in self._scores[submission.user]:
            task = task_store.retrieve(submission.task)
            self._scores[submission.user][submission.task] = \
                Score(score_mode=task["score_mode"])

        score_obj = self._scores[submission.user][submission.task]
        old_score = score_obj.get_score()
        score_obj.create_submission(key, submission)
        new_score = score_obj.get_score()
        if old_score != new_score:
            self.notify_callbacks(submission.user, submission.task, new_score)

    def update_submission(self, key, old_submission, submission):
        if old_submission.user != submission.user or \
                old_submission.task != submission.task:
            # TODO Delete all subchanges from the Score of the old
            # submission and create them on the new one.
            self.delete_submission(key, old_submission)
            self.create_submission(key, submission)
            return

        task = task_store.retrieve(submission.task)

        score_obj = self._scores[submission.user][submission.task]
        old_score = score_obj.get_score()
        score_obj.update_submission(key, submission)
        score_obj.update_score_mode(task["score_mode"])
        new_score = score_obj.get_score()
        if old_score != new_score:
            self.notify_callbacks(submission.user, submission.task, new_score)

    def delete_submission(self, key, submission):
        score_obj = self._scores[submission.user][submission.task]
        old_score = score_obj.get_score()
        score_obj.delete_submission(key)
        new_score = score_obj.get_score()
        if old_score != new_score:
            self.notify_callbacks(submission.user, submission.task, new_score)

        if len(self._scores[submission.user][submission.task]
               ._submissions) == 0:
            del self._scores[submission.user][submission.task]
        if len(self._scores[submission.user]) == 0:
            del self._scores[submission.user]

    def create_subchange(self, key, subchange):
        submission = submission_store._store[subchange.submission]
        score_obj = self._scores[submission.user][submission.task]
        old_score = score_obj.get_score()
        score_obj.create_subchange(key, subchange)
        new_score = score_obj.get_score()
        if old_score != new_score:
            self.notify_callbacks(submission.user, submission.task, new_score)

    def update_subchange(self, key, old_subchange, subchange):
        if old_subchange.submission != subchange.submission:
            self.delete_subchange(key, old_subchange)
            self.create_subchange(key, subchange)
            return

        submission = submission_store._store[subchange.submission]
        score_obj = self._scores[submission.user][submission.task]
        old_score = score_obj.get_score()
        score_obj.update_subchange(key, subchange)
        new_score = score_obj.get_score()
        if old_score != new_score:
            self.notify_callbacks(submission.user, submission.task, new_score)

    def delete_subchange(self, key, subchange):
        if subchange.submission not in submission_store:
            # Submission has just been deleted. We cannot retrieve the
            # user and the task, so we cannot clean up the Score obj.
            # But the delete_submission callback will do it for us!
            return
        submission = submission_store._store[subchange.submission]
        score_obj = self._scores[submission.user][submission.task]
        old_score = score_obj.get_score()
        score_obj.delete_subchange(key)
        new_score = score_obj.get_score()
        if old_score != new_score:
            self.notify_callbacks(submission.user, submission.task, new_score)

    def get_score(self, user, task):
        if user not in self._scores or task not in self._scores[user]:
            # We may want to raise an exception to distinguish between
            # "no submissions" and "submission with 0 points"
            return 0
        return self._scores[user][task].get_score()

    def get_submissions(self, user, task):
        if user not in self._scores or task not in self._scores[user]:
            return dict()
        return self._scores[user][task]._submissions

    def get_global_history(self):
        """Merge all individual histories into a global one.

        Take all per-user/per-task histories and merge them, providing
        a global history of all schore changes and return it using a
        generator.  Returned data is in the form (user_id, task_id,
        time, score).

        """
        # Use a priority queue, containing only one entry
        # per-user/per-task.
        queue = list()
        for user, dic in self._scores.iteritems():
            for task, scoring in dic.iteritems():
                if scoring._history:
                    heapq.heappush(queue, (scoring._history[0],
                                           user, task, scoring, 0))

        # When an entry is popped, push the next entry for that
        # user/task (if any).
        while len(queue) != 0:
            (time, score), user, task, scoring, index = heapq.heappop(queue)
            yield (user, task, time, score)
            if len(scoring._history) > index + 1:
                heapq.heappush(queue, (scoring._history[index + 1],
                                       user, task, scoring, index + 1))


store = ScoringStore()



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

# Contest Management System - http://cms-dev.github.io/
# Copyright © 2011-2013 Luca Wehrstedt <luca.wehrstedt@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals


class InvalidKey(Exception):
    """Exception raised in case of invalid key."""
    pass


class InvalidData(Exception):
    """Exception raised in case of invalid data."""
    pass


class Entity(object):
    """Base virtual class which all entities should extend.

    Provide some virtual methods that other classes should implement.

    """
    def set(self, data):
        """Set all properties using the given data.

        Accept the data format used on the HTTP interface.

        data (dict): the properties of the entity, in the "external"
            format

        Raise InvalidData if not able to parse the data argument.

        """
        pass

    def get(self):
        """Get all properties.

        Produce the data format used on the HTTP interface.

        return (dict): the properties of the entity, in the "external"
            format

        """
        pass

    def consistent(self):
        """Check if the entity is consistent.

        Verify that all references to other entities are correct (i.e.
        those entities actually exist).

        return (bool): the result of this check

        """
        return True



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

# Contest Management System - http://cms-dev.github.io/
# Copyright © 2011-2015 Luca Wehrstedt <luca.wehrstedt@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import io
import json
import logging
import os
import re

from gevent.lock import RLock

from cmsranking.Config import config
from cmsranking.Entity import Entity, InvalidKey, InvalidData


logger = logging.getLogger(__name__)


# Global shared lock for all Store instances.
LOCK = RLock()


class Store(object):
    """A store for entities.

    Provide methods to perform the CRUD operations (create, retrieve,
    update, delete) on a set of entities, accessed by their unique key.
    It's very similar to a dict, except that keys are strings, values
    are of a single type (defined at init-time) and it's possible to
    get notified when something changes by providing appropriate
    callbacks.

    """
    def __init__(self, entity, dir_name, depends=None):
        """Initialize an empty EntityStore.

        The entity definition given as argument will define what kind
        of entities will be stored. It cannot be changed.

        entity (type): the class definition of the entities that will
            be stored

        """
        if not issubclass(entity, Entity):
            raise ValueError("The 'entity' parameter "
                             "isn't a subclass of Entity")
        self._entity = entity
        self._path = os.path.join(config.lib_dir, dir_name)
        self._depends = depends if depends is not None else []
        self._store = dict()
        self._create_callbacks = list()
        self._update_callbacks = list()
        self._delete_callbacks = list()

    def load_from_disk(self):
        """Load the initial data for this store from the disk.

        """
        try:
            os.mkdir(self._path)
        except OSError:
            # it's ok: it means the directory already exists
            pass

        try:
            for name in os.listdir(self._path):
                # TODO check that the key is '[A-Za-z0-9_]+'
                if name[-5:] == '.json' and name[:-5] != '':
                    with io.open(os.path.join(self._path, name), 'rb') as rec:
                        item = self._entity()
                        item.set(json.load(rec, encoding='utf-8'))
                        item.key = name[:-5]
                        self._store[name[:-5]] = item
        except OSError:
            # the path isn't a directory or is inaccessible
            logger.error("Path is not a directory or is not accessible",
                         exc_info=True)
        except IOError:
            logger.error("I/O error occured", exc_info=True)
        except ValueError:
            logger.error("Invalid JSON", exc_info=False,
                         extra={'location': os.path.join(self._path, name)})
        except InvalidData as exc:
            logger.error(exc.message, exc_info=False,
                         extra={'location': os.path.join(self._path, name)})

    def add_create_callback(self, callback):
        """Add a callback to be called when entities are created.

        Callbacks can be any kind of callable objects. They must accept
        a single argument: the key of the entity.

        """
        self._create_callbacks.append(callback)

    def add_update_callback(self, callback):
        """Add a callback to be called when entities are updated.

        Callbacks can be any kind of callable objects. They must accept
        a single argument: the key of the entity.

        """
        self._update_callbacks.append(callback)

    def add_delete_callback(self, callback):
        """Add a callback to be called when entities are deleted.

        Callbacks can be any kind of callable objects. They must accept
        a single argument: the key of the entity.

        """
        self._delete_callbacks.append(callback)

    def create(self, key, data):
        """Create a new entity.

        Create a new entity with the given key and the given data.

        key (unicode): the key with which the entity will be later
            accessed
        data (dict): the properties of the entity

        raise (InvalidKey): if key isn't a unicode or if an entity
            with the same key is already present in the store.
        raise (InvalidData): if data cannot be parsed, if it's missing
            some properties or if properties are of the wrong type.

        """
        if not isinstance(key, unicode) or key in self._store:
            raise InvalidKey("Key already in store.")

        # create entity
        with LOCK:
            item = self._entity()
            item.set(data)
            if not item.consistent():
                raise InvalidData("Inconsistent data")
            item.key = key
            self._store[key] = item
            # notify callbacks
            for callback in self._create_callbacks:
                callback(key, item)
            # reflect changes on the persistent storage
            try:
                path = os.path.join(self._path, key + '.json')
                with io.open(path, 'wb') as rec:
                    json.dump(self._store[key].get(), rec, encoding='utf-8')
            except IOError:
                logger.error("I/O error occured while creating entity",
                             exc_info=True)

    def update(self, key, data):
        """Update an entity.

        Update an existing entity with the given key and the given
        data.

        key (unicode): the key of the entity that has to be updated
        data (dict): the new properties of the entity

        raise (InvalidKey): if key isn't a unicode or if no entity
            with that key is present in the store.
        raise (InvalidData): if data cannot be parsed, if it's missing
            some properties or if properties are of the wrong type.

        """
        if not isinstance(key, unicode) or key not in self._store:
            raise InvalidKey("Key not in store.")

        # update entity
        with LOCK:
            item = self._entity()
            item.set(data)
            if not item.consistent():
                raise InvalidData("Inconsistent data")
            item.key = key
            old_item = self._store[key]
            self._store[key] = item
            # notify callbacks
            for callback in self._update_callbacks:
                callback(key, old_item, item)
            # reflect changes on the persistent storage
            try:
                path = os.path.join(self._path, key + '.json')
                with io.open(path, 'wb') as rec:
                    json.dump(self._store[key].get(), rec, encoding='utf-8')
            except IOError:
                logger.error("I/O error occured while updating entity",
                             exc_info=True)

    def merge_list(self, data_dict):
        """Merge a list of entities.

        Take a dictionary of entites and, for each of them:
         - if it's not present in the store, create it
         - if it's present, update it

        data_dict (dict): the dictionary of entities

        raise (InvalidData) if data cannot be parsed, if an entity is
            missing some properties or if properties are of the wrong
            type.

        """
        with LOCK:
            if not isinstance(data_dict, dict):
                raise InvalidData("Not a dictionary")
            item_dict = dict()
            for key, value in data_dict.iteritems():
                try:
                    # FIXME We should allow keys to be arbitrary unicode
                    # strings, so this just needs to be a non-empty check.
                    if not re.match('[A-Za-z0-9_]+', key):
                        raise InvalidData("Invalid key")
                    item = self._entity()
                    item.set(value)
                    if not item.consistent():
                        raise InvalidData("Inconsistent data")
                    item.key = key
                    item_dict[key] = item
                except InvalidData as exc:
                    exc.message = "[entity %s] %s" % (key, exc)
                    exc.args = exc.message,
                    raise exc

            for key, value in item_dict.iteritems():
                is_new = key not in self._store
                old_value = self._store.get(key)
                # insert entity
                self._store[key] = value
                # notify callbacks
                if is_new:
                    for callback in self._create_callbacks:
                        callback(key, value)
                else:
                    for callback in self._update_callbacks:
                        callback(key, old_value, value)
                # reflect changes on the persistent storage
                try:
                    path = os.path.join(self._path, key + '.json')
                    with io.open(path, 'wb') as rec:
                        json.dump(value.get(), rec, encoding='utf-8')
                except IOError:
                    logger.error(
                        "I/O error occured while merging entity lists",
                        exc_info=True)

    def delete(self, key):
        """Delete an entity.

        Delete an existing entity from the store.

        key (unicode): the key of the entity that has to be deleted

        raise (InvalidKey): if key isn't a unicode or if no entity
            with that key is present in the store.

        """
        if not isinstance(key, unicode) or key not in self._store:
            raise InvalidKey("Key not in store.")

        with LOCK:
            # delete entity
            old_value = self._store[key]
            del self._store[key]
            # enforce consistency
            for depend in self._depends:
                for o_key, o_value in list(depend._store.iteritems()):
                    if not o_value.consistent():
                        depend.delete(o_key)
            # notify callbacks
            for callback in self._delete_callbacks:
                callback(key, old_value)
            # reflect changes on the persistent storage
            try:
                os.remove(os.path.join(self._path, key + '.json'))
            except OSError:
                logger.error("Unable to delete entity", exc_info=True)

    def delete_list(self):
        """Delete all entities.

        Delete all existing entities from the store.

        """
        with LOCK:
            # delete all entities
            for key in list(self._store.iterkeys()):
                self.delete(key)

    def retrieve(self, key):
        """Retrieve an entity.

        Retrieve an existing entity from the store.

        key (unicode): the key of the entity that has to be retrieved

        raise (InvalidKey): if key isn't a unicode or if no entity
            with that key is present in the store.

        """
        if not isinstance(key, unicode) or key not in self._store:
            raise InvalidKey("Key not in store.")

        # retrieve entity
        return self._store[key].get()

    def retrieve_list(self):
        """Retrieve a list of all entities."""
        result = dict()
        for key, value in self._store.iteritems():
            result[key] = value.get()
        return result

    def __contains__(self, key):
        return key in self._store



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

# Contest Management System - http://cms-dev.github.io/
# Copyright © 2011-2013 Luca Wehrstedt <luca.wehrstedt@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import six

from cmsranking.Entity import Entity, InvalidData
from cmsranking.Store import Store
from cmsranking.Task import store as task_store


class Contest(Entity):
    """The entity representing a contest.

    It consists of the following properties:
    - name (unicode): the human-readable name of the contest
    - begin (int): the unix timestamp at which the contest begins
    - end (int): the unix timestamp at which the contest ends
    - score_precision (int): how many decimal places to show in scores

    """
    def __init__(self):
        """Set the properties to some default values.

        """
        Entity.__init__(self)
        self.name = None
        self.begin = None
        self.end = None
        self.score_precision = None

    @staticmethod
    def validate(data):
        """Validate the given dictionary.

        See if it contains a valid representation of this entity.

        """
        try:
            assert isinstance(data, dict), \
                "Not a dictionary"
            assert isinstance(data['name'], six.text_type), \
                "Field 'name' isn't a string"
            assert isinstance(data['begin'], six.integer_types), \
                "Field 'begin' isn't an integer"
            assert isinstance(data['end'], six.integer_types), \
                "Field 'end' isn't an integer"
            assert data['begin'] <= data['end'], \
                "Field 'begin' is greater than 'end'"
            assert isinstance(data['score_precision'], six.integer_types), \
                "Field 'score_precision' isn't an integer"
            assert data['score_precision'] >= 0, \
                "Field 'score_precision' is negative"
        except KeyError as exc:
            raise InvalidData("Field %s is missing" % exc.message)
        except AssertionError as exc:
            raise InvalidData(exc.message)

    def set(self, data):
        self.validate(data)
        self.name = data['name']
        self.begin = data['begin']
        self.end = data['end']
        self.score_precision = data['score_precision']

    def get(self):
        result = self.__dict__.copy()
        del result['key']
        return result


store = Store(Contest, 'contests', [task_store])



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

# Contest Management System - http://cms-dev.github.io/
# Copyright © 2011-2013 Luca Wehrstedt <luca.wehrstedt@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import six

from cmsranking.Entity import Entity, InvalidData
from cmsranking.Store import Store
from cmsranking.Subchange import store as subchange_store


class Submission(Entity):
    """The entity representing a submission.

    It consists of the following properties:
    - user (unicode): the key of the user who submitted
    - task (unicode): the key of the task of the submission
    - time (int): the time the submission has been submitted

    """
    def __init__(self):
        """Set the properties to some default values.

        """
        Entity.__init__(self)
        self.user = None
        self.task = None
        self.time = None

    @staticmethod
    def validate(data):
        """Validate the given dictionary.

        See if it contains a valid representation of this entity.

        """
        try:
            assert isinstance(data, dict), \
                "Not a dictionary"
            assert isinstance(data['user'], six.text_type), \
                "Field 'user' isn't a string"
            assert isinstance(data['task'], six.text_type), \
                "Field 'task' isn't a string"
            assert isinstance(data['time'], six.integer_types), \
                "Field 'time' isn't an integer (unix timestamp)"
        except KeyError as exc:
            raise InvalidData("Field %s is missing" % exc.message)
        except AssertionError as exc:
            raise InvalidData(exc.message)

    def set(self, data):
        self.validate(data)
        self.user = data['user']
        self.task = data['task']
        self.time = data['time']

    def get(self):
        result = self.__dict__.copy()
        del result['key']
        del result['score']
        del result['token']
        del result['extra']
        return result

    def consistent(self):
        from cmsranking.Task import store as task_store
        from cmsranking.User import store as user_store
        return self.task in task_store and self.user in user_store


store = Store(Submission, 'submissions', [subchange_store])



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

# Contest Management System - http://cms-dev.github.io/
# Copyright © 2011-2015 Luca Wehrstedt <luca.wehrstedt@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import argparse
import functools
import io
import json
import logging
import os
import pprint
import re
import shutil
import time
from datetime import datetime

import gevent
from gevent.pywsgi import WSGIServer

from werkzeug.wrappers import Request, Response
from werkzeug.routing import Map, Rule
from werkzeug.exceptions import HTTPException, BadRequest, Unauthorized, \
    Forbidden, NotFound, NotAcceptable, UnsupportedMediaType
from werkzeug.wsgi import responder, wrap_file, SharedDataMiddleware, \
    DispatcherMiddleware
from werkzeug.utils import redirect

# Needed for initialization. Do not remove.
import cmsranking.Logger

from cmscommon.eventsource import EventSource
from cmsranking.Config import config
from cmsranking.Entity import InvalidData
import cmsranking.Contest as Contest
import cmsranking.Task as Task
import cmsranking.Team as Team
import cmsranking.User as User
import cmsranking.Submission as Submission
import cmsranking.Subchange as Subchange
import cmsranking.Scoring as Scoring


logger = logging.getLogger(__name__)


class CustomUnauthorized(Unauthorized):
    def get_response(self, environ=None):
        response = Unauthorized.get_response(self, environ)
        # XXX With werkzeug-0.9 a full-featured Response object is
        # returned: there is no need for this.
        response = Response.force_type(response)
        response.www_authenticate.set_basic(config.realm_name)
        return response


class StoreHandler(object):
    def __init__(self, store):
        self.store = store

        self.router = Map(
            [Rule("/<key>", methods=["GET"], endpoint="get"),
             Rule("/", methods=["GET"], endpoint="get_list"),
             Rule("/<key>", methods=["PUT"], endpoint="put"),
             Rule("/", methods=["PUT"], endpoint="put_list"),
             Rule("/<key>", methods=["DELETE"], endpoint="delete"),
             Rule("/", methods=["DELETE"], endpoint="delete_list"),
             ], encoding_errors="strict")

    def __call__(self, environ, start_response):
        return self.wsgi_app(environ, start_response)

    @responder
    def wsgi_app(self, environ, start_response):
        route = self.router.bind_to_environ(environ)
        try:
            endpoint, args = route.match()
        except HTTPException as exc:
            return exc

        request = Request(environ)
        request.encoding_errors = "strict"

        response = Response()

        try:
            if endpoint == "get":
                self.get(request, response, args["key"])
            elif endpoint == "get_list":
                self.get_list(request, response)
            elif endpoint == "put":
                self.put(request, response, args["key"])
            elif endpoint == "put_list":
                self.put_list(request, response)
            elif endpoint == "delete":
                self.delete(request, response, args["key"])
            elif endpoint == "delete_list":
                self.delete_list(request, response)
            else:
                raise RuntimeError()
        except HTTPException as exc:
            return exc

        return response

    def authorized(self, request):
        return request.authorization is not None and \
            request.authorization.type == "basic" and \
            request.authorization.username == config.username and \
            request.authorization.password == config.password

    def get(self, request, response, key):
        # Limit charset of keys.
        if re.match("^[A-Za-z0-9_]+$", key) is None:
            return NotFound()
        if key not in self.store:
            raise NotFound()
        if request.accept_mimetypes.quality("application/json") <= 0:
            raise NotAcceptable()

        response.status_code = 200
        response.headers[b'Timestamp'] = b"%0.6f" % time.time()
        response.mimetype = "application/json"
        response.data = json.dumps(self.store.retrieve(key))

    def get_list(self, request, response):
        if request.accept_mimetypes.quality("application/json") <= 0:
            raise NotAcceptable()

        response.status_code = 200
        response.headers[b'Timestamp'] = b"%0.6f" % time.time()
        response.mimetype = "application/json"
        response.data = json.dumps(self.store.retrieve_list())

    def put(self, request, response, key):
        # Limit charset of keys.
        if re.match("^[A-Za-z0-9_]+$", key) is None:
            return Forbidden()
        if not self.authorized(request):
            logger.warning("Unauthorized request.",
                           extra={'location': request.url,
                                  'details': repr(request.authorization)})
            raise CustomUnauthorized()
        if request.mimetype != "application/json":
            logger.warning("Unsupported MIME type.",
                           extra={'location': request.url,
                                  'details': request.mimetype})
            raise UnsupportedMediaType()

        try:
            data = json.load(request.stream)
        except (TypeError, ValueError):
            logger.warning("Wrong JSON.",
                           extra={'location': request.url})
            raise BadRequest()

        try:
            if key not in self.store:
                self.store.create(key, data)
            else:
                self.store.update(key, data)
        except InvalidData:
            logger.warning("Invalid data.", exc_info=True,
                           extra={'location': request.url,
                                  'details': pprint.pformat(data)})
            raise BadRequest()

        response.status_code = 204

    def put_list(self, request, response):
        if not self.authorized(request):
            logger.info("Unauthorized request.",
                        extra={'location': request.url,
                               'details': repr(request.authorization)})
            raise CustomUnauthorized()
        if request.mimetype != "application/json":
            logger.warning("Unsupported MIME type.",
                           extra={'location': request.url,
                                  'details': request.mimetype})
            raise UnsupportedMediaType()

        try:
            data = json.load(request.stream)
        except (TypeError, ValueError):
            logger.warning("Wrong JSON.",
                           extra={'location': request.url})
            raise BadRequest()

        try:
            self.store.merge_list(data)
        except InvalidData:
            logger.warning("Invalid data.", exc_info=True,
                           extra={'location': request.url,
                                  'details': pprint.pformat(data)})
            raise BadRequest()

        response.status_code = 204

    def delete(self, request, response, key):
        # Limit charset of keys.
        if re.match("^[A-Za-z0-9_]+$", key) is None:
            return NotFound()
        if key not in self.store:
            raise NotFound()
        if not self.authorized(request):
            logger.info("Unauthorized request.",
                        extra={'location': request.url,
                               'details': repr(request.authorization)})
            raise CustomUnauthorized()

        self.store.delete(key)

        response.status_code = 204

    def delete_list(self, request, response):
        if not self.authorized(request):
            logger.info("Unauthorized request.",
                        extra={'location': request.url,
                               'details': repr(request.authorization)})
            raise CustomUnauthorized()

        self.store.delete_list()

        response.status_code = 204


class DataWatcher(EventSource):
    """Receive the messages from the entities store and redirect them."""
    def __init__(self):
        self._CACHE_SIZE = config.buffer_size
        EventSource.__init__(self)

        Contest.store.add_create_callback(
            functools.partial(self.callback, "contest", "create"))
        Contest.store.add_update_callback(
            functools.partial(self.callback, "contest", "update"))
        Contest.store.add_delete_callback(
            functools.partial(self.callback, "contest", "delete"))

        Task.store.add_create_callback(
            functools.partial(self.callback, "task", "create"))
        Task.store.add_update_callback(
            functools.partial(self.callback, "task", "update"))
        Task.store.add_delete_callback(
            functools.partial(self.callback, "task", "delete"))

        Team.store.add_create_callback(
            functools.partial(self.callback, "team", "create"))
        Team.store.add_update_callback(
            functools.partial(self.callback, "team", "update"))
        Team.store.add_delete_callback(
            functools.partial(self.callback, "team", "delete"))

        User.store.add_create_callback(
            functools.partial(self.callback, "user", "create"))
        User.store.add_update_callback(
            functools.partial(self.callback, "user", "update"))
        User.store.add_delete_callback(
            functools.partial(self.callback, "user", "delete"))

        Scoring.store.add_score_callback(self.score_callback)

    def callback(self, entity, event, key, *args):
        self.send(entity, "%s %s" % (event, key))

    def score_callback(self, user, task, score):
        # FIXME Use score_precision.
        self.send("score", "%s %s %0.2f" % (user, task, score))


def SubListHandler(request, response, user_id):
    if request.accept_mimetypes.quality("application/json") <= 0:
        raise NotAcceptable()

    result = list()
    for task_id in Task.store._store.iterkeys():
        result.extend(Scoring.store.get_submissions(user_id, task_id).values())
    result.sort(key=lambda x: (x.task, x.time))
    result = list(a.__dict__ for a in result)

    response.status_code = 200
    response.mimetype = "application/json"
    response.data = json.dumps(result)


def HistoryHandler(request, response):
    if request.accept_mimetypes.quality("application/json") <= 0:
        raise NotAcceptable()

    result = list(Scoring.store.get_global_history())

    response.status_code = 200
    response.mimetype = "application/json"
    response.data = json.dumps(result)


def ScoreHandler(request, response):
    if request.accept_mimetypes.quality("application/json") <= 0:
        raise NotAcceptable()

    result = dict()
    for u_id, tasks in Scoring.store._scores.iteritems():
        for t_id, score in tasks.iteritems():
            if score.get_score() > 0.0:
                result.setdefault(u_id, dict())[t_id] = score.get_score()

    response.status_code = 200
    response.headers[b'Timestamp'] = b"%0.6f" % time.time()
    response.mimetype = "application/json"
    response.data = json.dumps(result)


class ImageHandler(object):
    EXT_TO_MIME = {
        'png': 'image/png',
        'jpg': 'image/jpeg',
        'gif': 'image/gif',
        'bmp': 'image/bmp'
    }

    MIME_TO_EXT = dict((v, k) for k, v in EXT_TO_MIME.iteritems())

    def __init__(self, location, fallback):
        self.location = location
        self.fallback = fallback

        self.router = Map(
            [Rule("/<name>", methods=["GET"], endpoint="get"),
             ], encoding_errors="strict")

    def __call__(self, environ, start_response):
        return self.wsgi_app(environ, start_response)

    @responder
    def wsgi_app(self, environ, start_response):
        route = self.router.bind_to_environ(environ)
        try:
            endpoint, args = route.match()
        except HTTPException as exc:
            return exc

        location = self.location % args

        request = Request(environ)
        request.encoding_errors = "strict"

        response = Response()

        available = list()
        for extension, mimetype in self.EXT_TO_MIME.iteritems():
            if os.path.isfile(location + '.' + extension):
                available.append(mimetype)
        mimetype = request.accept_mimetypes.best_match(available)
        if mimetype is not None:
            path = "%s.%s" % (location, self.MIME_TO_EXT[mimetype])
        else:
            path = self.fallback
            mimetype = 'image/png'  # FIXME Hardcoded type.

        response.status_code = 200
        response.mimetype = mimetype
        response.last_modified = \
            datetime.utcfromtimestamp(os.path.getmtime(path))\
                    .replace(microsecond=0)

        # TODO check for If-Modified-Since and If-None-Match

        response.response = wrap_file(environ, io.open(path, 'rb'))
        response.direct_passthrough = True

        return response


class RoutingHandler(object):
    def __init__(self, event_handler, logo_handler):
        self.router = Map(
            [Rule("/", methods=["GET"], endpoint="root"),
             Rule("/sublist/<user_id>", methods=["GET"], endpoint="sublist"),
             Rule("/history", methods=["GET"], endpoint="history"),
             Rule("/scores", methods=["GET"], endpoint="scores"),
             Rule("/events", methods=["GET"], endpoint="events"),
             Rule("/logo", methods=["GET"], endpoint="logo"),
             ], encoding_errors="strict")

        self.event_handler = event_handler
        self.logo_handler = logo_handler
        self.root_handler = redirect("Ranking.html")

    def __call__(self, environ, start_response):
        return self.wsgi_app(environ, start_response)

    def wsgi_app(self, environ, start_response):
        route = self.router.bind_to_environ(environ)
        try:
            endpoint, args = route.match()
        except HTTPException as exc:
            return exc(environ, start_response)

        if endpoint == "events":
            return self.event_handler(environ, start_response)
        elif endpoint == "logo":
            return self.logo_handler(environ, start_response)
        elif endpoint == "root":
            return self.root_handler(environ, start_response)
        else:
            request = Request(environ)
            request.encoding_errors = "strict"

            response = Response()

            if endpoint == "sublist":
                SubListHandler(request, response, args["user_id"])
            elif endpoint == "scores":
                ScoreHandler(request, response)
            elif endpoint == "history":
                HistoryHandler(request, response)

            return response(environ, start_response)


def main():
    """Entry point for RWS.

    return (bool): True if executed successfully.

    """
    parser = argparse.ArgumentParser(
        description="Ranking for CMS.")
    parser.add_argument("-d", "--drop", action="store_true",
                        help="drop the data already stored")
    args = parser.parse_args()

    if args.drop:
        print("Are you sure you want to delete directory %s? [y/N]" %
              config.lib_dir, end='')
        ans = raw_input().lower()
        if ans in ['y', 'yes']:
            print("Removing directory %s." % config.lib_dir)
            shutil.rmtree(config.lib_dir)
        else:
            print("Not removing directory %s." % config.lib_dir)
        return False

    Contest.store.load_from_disk()
    Task.store.load_from_disk()
    Team.store.load_from_disk()
    User.store.load_from_disk()
    Submission.store.load_from_disk()
    Subchange.store.load_from_disk()

    Scoring.store.init_store()

    toplevel_handler = RoutingHandler(DataWatcher(), ImageHandler(
        os.path.join(config.lib_dir, '%(name)s'),
        os.path.join(config.web_dir, 'img', 'logo.png')))

    wsgi_app = SharedDataMiddleware(DispatcherMiddleware(
        toplevel_handler,
        {'/contests': StoreHandler(Contest.store),
         '/tasks': StoreHandler(Task.store),
         '/teams': StoreHandler(Team.store),
         '/users': StoreHandler(User.store),
         '/submissions': StoreHandler(Submission.store),
         '/subchanges': StoreHandler(Subchange.store),
         '/faces': ImageHandler(
             os.path.join(config.lib_dir, 'faces', '%(name)s'),
             os.path.join(config.web_dir, 'img', 'face.png')),
         '/flags': ImageHandler(
             os.path.join(config.lib_dir, 'flags', '%(name)s'),
             os.path.join(config.web_dir, 'img', 'flag.png')),
         }), {'/': config.web_dir})

    servers = list()
    if config.http_port is not None:
        http_server = WSGIServer(
            (config.bind_address, config.http_port), wsgi_app)
        servers.append(http_server)
    if config.https_port is not None:
        https_server = WSGIServer(
            (config.bind_address, config.https_port), wsgi_app,
            certfile=config.https_certfile, keyfile=config.https_keyfile)
        servers.append(https_server)

    try:
        gevent.joinall(list(gevent.spawn(s.serve_forever) for s in servers))
    except KeyboardInterrupt:
        pass
    finally:
        gevent.joinall(list(gevent.spawn(s.stop) for s in servers))
    return True



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

# Contest Management System - http://cms-dev.github.io/
# Copyright © 2011-2013 Luca Wehrstedt <luca.wehrstedt@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import six

from cmsranking.Entity import Entity, InvalidData
from cmsranking.Store import Store
from cmsranking.Submission import store as submission_store


class Task(Entity):
    """The entity representing a task.

    It consists of the following properties:
    - name (unicode): the human-readable name of the task
    - short_name (unicode): a shorter name for the task, usually a
        code-name
    - contest (unicode): the id of the contest the task belongs to
    - max_score (float): the maximum achievable score for the task
    - score_precision (int): how many decimal places to show in scores
    - data_headers ([unicode]): a list with the descriptions of the
        extra fields that will be provided with each submission for the
        task
    - order (int): the order of the tasks inside of the contest

    """
    def __init__(self):
        """Set the properties to some default values.

        """
        Entity.__init__(self)
        self.name = None
        self.short_name = None
        self.contest = None
        self.max_score = None
        self.extra_headers = None
        self.order = None
        self.score_mode = "max_tokened_last"

    @staticmethod
    def validate(data):
        """Validate the given dictionary.

        See if it contains a valid representation of this entity.

        """
        try:
            assert isinstance(data, dict), \
                "Not a dictionary"
            assert isinstance(data['name'], six.text_type), \
                "Field 'name' isn't a string"
            assert isinstance(data['short_name'], six.text_type), \
                "Field 'short_name' isn't a string"
            assert isinstance(data['contest'], six.text_type), \
                "Field 'contest' isn't a string"
            assert isinstance(data['max_score'], float), \
                "Field 'max_score' isn't a float"
            assert isinstance(data['score_precision'], six.integer_types), \
                "Field 'score_precision' isn't an integer"
            assert data['score_precision'] >= 0, \
                "Field 'score_precision' is negative"
            assert isinstance(data['extra_headers'], list), \
                "Field 'extra_headers' isn't a list of strings"
            assert isinstance(data['score_mode'], six.text_type), \
                "Field 'score_mode' isn't a string"
            for i in data['extra_headers']:
                assert isinstance(i, six.text_type), \
                    "Field 'extra_headers' isn't a list of strings"
            assert isinstance(data['order'], six.integer_types), \
                "Field 'order' isn't an integer"
        except KeyError as exc:
            raise InvalidData("Field %s is missing" % exc.message)
        except AssertionError as exc:
            raise InvalidData(exc.message)

    def set(self, data):
        self.validate(data)
        self.name = data['name']
        self.short_name = data['short_name']
        self.contest = data['contest']
        self.max_score = data['max_score']
        self.score_precision = data['score_precision']
        self.extra_headers = data['extra_headers']
        self.order = data['order']
        self.score_mode = data['score_mode']

    def get(self):
        result = self.__dict__.copy()
        del result['key']
        return result

    def consistent(self):
        from cmsranking.Contest import store as contest_store
        return self.contest in contest_store


store = Store(Task, 'tasks', [submission_store])



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

# Contest Management System - http://cms-dev.github.io/
# Copyright © 2011-2013 Luca Wehrstedt <luca.wehrstedt@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import six

from cmsranking.Entity import Entity, InvalidData
from cmsranking.Store import Store
from cmsranking.Submission import store as submission_store


class User(Entity):
    """The entity representing a user.

    It consists of the following properties:
    - f_name (unicode): the first name of the user
    - l_name (unicode): the last name of the user
    - team (unicode): the id of the team the user belongs to

    """
    def __init__(self):
        """Set the properties to some default values.

        """
        Entity.__init__(self)
        self.f_name = None
        self.l_name = None
        self.team = None

    @staticmethod
    def validate(data):
        """Validate the given dictionary.

        See if it contains a valid representation of this entity.

        """
        try:
            assert isinstance(data, dict), \
                "Not a dictionary"
            assert isinstance(data['f_name'], six.text_type), \
                "Field 'f_name' isn't a string"
            assert isinstance(data['l_name'], six.text_type), \
                "Field 'l_name' isn't a string"
            assert data['team'] is None or \
                isinstance(data['team'], six.text_type), \
                "Field 'team' isn't a string (or null)"
        except KeyError as exc:
            raise InvalidData("Field %s is missing" % exc.message)
        except AssertionError as exc:
            raise InvalidData(exc.message)

    def set(self, data):
        self.validate(data)
        self.f_name = data['f_name']
        self.l_name = data['l_name']
        self.team = data['team']

    def get(self):
        result = self.__dict__.copy()
        del result['key']
        return result

    def consistent(self):
        from cmsranking.Team import store as team_store
        return self.team is None or self.team in team_store


store = Store(User, 'users', [submission_store])



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

# Contest Management System - http://cms-dev.github.io/
# Copyright © 2011-2013 Luca Wehrstedt <luca.wehrstedt@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import curses
import logging
import os.path
import sys
import time
from traceback import format_tb

import gevent.coros

from cmsranking.Config import config


class StreamHandler(logging.StreamHandler):
    """Subclass to make gevent-aware.

    Use a gevent lock instead of a threading one to block only the
    current greenlet.

    """
    def createLock(self):
        """Set self.lock to a new gevent RLock.

        """
        self.lock = gevent.coros.RLock()


class FileHandler(logging.FileHandler):
    """Subclass to make gevent-aware.

    Use a gevent lock instead of a threading one to block only the
    current greenlet.

    """
    def createLock(self):
        """Set self.lock to a new gevent RLock.

        """
        self.lock = gevent.coros.RLock()


def has_color_support(stream):
    """Try to determine if the given stream supports colored output.

    Return True only if the stream declares to be a TTY, if it has a
    file descriptor on which ncurses can initialize a terminal and if
    that terminal's entry in terminfo declares support for colors.

    stream (fileobj): a file-like object (that adheres to the API
        declared in the `io' package).

    return (bool): True if we're sure that colors are supported, False
        if they aren't or if we can't tell.

    """
    if stream.isatty():
        try:
            curses.setupterm(fd=stream.fileno())
            # See `man terminfo` for capabilities' names and meanings.
            if curses.tigetnum("colors") > 0:
                return True
        # fileno() can raise IOError or OSError (since Python 3.3).
        except Exception:
            pass
    return False


## ANSI utilities. See for reference:
# http://pueblo.sourceforge.net/doc/manual/ansi_color_codes.html
# http://en.wikipedia.org/wiki/ANSI_escape_code
#

ANSI_FG_COLORS = {'black': 30,
                  'red': 31,
                  'green': 32,
                  'yellow': 33,
                  'blue': 34,
                  'magenta': 35,
                  'cyan': 36,
                  'white': 37}

ANSI_BG_COLORS = {'black': 40,
                  'red': 41,
                  'green': 42,
                  'yellow': 43,
                  'blue': 44,
                  'magenta': 45,
                  'cyan': 46,
                  'white': 47}

ANSI_RESET_CMD = 0
ANSI_FG_DEFAULT_CMD = 39
ANSI_BG_DEFAULT_CMD = 49
ANSI_BOLD_ON_CMD = 1
ANSI_BOLD_OFF_CMD = 22
ANSI_FAINT_ON_CMD = 2
ANSI_FAINT_OFF_CMD = 22
ANSI_ITALICS_ON_CMD = 3
ANSI_ITALICS_OFF_CMD = 23
ANSI_UNDERLINE_ON_CMD = 4
ANSI_UNDERLINE_OFF_CMD = 24
ANSI_STRIKETHROUGH_ON_CMD = 9
ANSI_STRIKETHROUGH_OFF_CMD = 29
ANSI_INVERSE_ON_CMD = 7
ANSI_INVERSE_OFF_CMD = 27

# TODO missing:
# - distinction between single and double underline
# - "slow blink on", "rapid blink on" and "blink off"
# - "conceal on" and "conceal off" (also called reveal)


class CustomFormatter(logging.Formatter):
    """A custom Formatter for our logs.

    """
    def __init__(self, color=True, *args, **kwargs):
        """Initialize the formatter.

        Based on the 'color' parameter we set the tags for many
        elements of our formatted output.

        """
        logging.Formatter.__init__(self, *args, **kwargs)

        self.color = color

        self.time_prefix = self.ansi_command(ANSI_BOLD_ON_CMD)
        self.time_suffix = self.ansi_command(ANSI_BOLD_OFF_CMD)
        self.cri_prefix = self.ansi_command(ANSI_BOLD_ON_CMD,
                                            ANSI_FG_COLORS['white'],
                                            ANSI_BG_COLORS['red'])
        self.cri_suffix = self.ansi_command(ANSI_BOLD_OFF_CMD,
                                            ANSI_FG_DEFAULT_CMD,
                                            ANSI_BG_DEFAULT_CMD)
        self.err_prefix = self.ansi_command(ANSI_BOLD_ON_CMD,
                                            ANSI_FG_COLORS['red'])
        self.err_suffix = self.ansi_command(ANSI_BOLD_OFF_CMD,
                                            ANSI_FG_DEFAULT_CMD)
        self.wrn_prefix = self.ansi_command(ANSI_BOLD_ON_CMD,
                                            ANSI_FG_COLORS['yellow'])
        self.wrn_suffix = self.ansi_command(ANSI_BOLD_OFF_CMD,
                                            ANSI_FG_DEFAULT_CMD)
        self.inf_prefix = self.ansi_command(ANSI_BOLD_ON_CMD,
                                            ANSI_FG_COLORS['green'])
        self.inf_suffix = self.ansi_command(ANSI_BOLD_OFF_CMD,
                                            ANSI_FG_DEFAULT_CMD)
        self.dbg_prefix = self.ansi_command(ANSI_BOLD_ON_CMD,
                                            ANSI_FG_COLORS['blue'])
        self.dbg_suffix = self.ansi_command(ANSI_BOLD_OFF_CMD,
                                            ANSI_FG_DEFAULT_CMD)

    def ansi_command(self, *args):
        """Produce the escape string that corresponds to the given
        ANSI command.

        """
        return "\033[%sm" % ';'.join(["%s" % x
                                      for x in args]) if self.color else ''

    def formatException(self, exc_info):
        exc_type, exc_value, traceback = exc_info

        result = "%sException%s %s.%s%s%s:\n\n    %s\n\n" % \
            (self.ansi_command(ANSI_BOLD_ON_CMD),
             self.ansi_command(ANSI_BOLD_OFF_CMD),
             exc_type.__module__,
             self.ansi_command(ANSI_BOLD_ON_CMD),
             exc_type.__name__,
             self.ansi_command(ANSI_BOLD_OFF_CMD),
             exc_value)
        result += "%sTraceback (most recent call last):%s\n%s" % \
            (self.ansi_command(ANSI_FAINT_ON_CMD),
             self.ansi_command(ANSI_FAINT_OFF_CMD),
             '\n'.join(map(lambda a: self.ansi_command(ANSI_FAINT_ON_CMD) +
                           a + self.ansi_command(ANSI_FAINT_OFF_CMD),
                           ''.join(format_tb(traceback)).strip().split('\n'))))

        return result

    def format(self, record):
        """Do the actual formatting.

        Prepend a timestamp and an abbreviation of the logging level,
        followed by the message, the request_body (if present) and the
        exception details (if present).

        """
        result = '%s%s.%03d%s' % \
            (self.time_prefix,
             self.formatTime(record, '%Y-%m-%d %H:%M:%S'), record.msecs,
             self.time_suffix)

        if record.levelno == logging.CRITICAL:
            result += ' %s CRI %s ' % (self.cri_prefix, self.cri_suffix)
        elif record.levelno == logging.ERROR:
            result += ' %s ERR %s ' % (self.err_prefix, self.err_suffix)
        elif record.levelno == logging.WARNING:
            result += ' %s WRN %s ' % (self.wrn_prefix, self.wrn_suffix)
        elif record.levelno == logging.INFO:
            result += ' %s INF %s ' % (self.inf_prefix, self.inf_suffix)
        else:  # DEBUG
            result += ' %s DBG %s ' % (self.dbg_prefix, self.dbg_suffix)

        try:
            message = record.getMessage()
        except Exception, exc:
            message = 'Bad message (%r): %r' % (exc, record.__dict__)

        result += message.strip()

        if "location" in record.__dict__:
            result += "\n%s" % record.location.strip()

        if "details" in record.__dict__:
            result += "\n\n%s" % record.details.strip()

        if record.exc_info:
            result += "\n\n%s" % self.formatException(record.exc_info).strip()

        return result.replace("\n", "\n    ") + '\n'


# Create a global reference to the root logger.
root_logger = logging.getLogger()
# Catch all logging messages (we'll filter them on the handlers).
root_logger.setLevel(logging.DEBUG)

# Define the stream handler to output on stderr.
shell_handler = StreamHandler(sys.stdout)
shell_handler.setLevel(logging.INFO)
shell_handler.setFormatter(CustomFormatter(has_color_support(sys.stdout)))
root_logger.addHandler(shell_handler)

# Define the file handler to output on the specified log directory.
log_filename = time.strftime("%Y-%m-%d-%H-%M-%S.log")
file_handler = FileHandler(os.path.join(config.log_dir, log_filename),
                           mode='w', encoding='utf-8')
file_handler.setLevel(logging.INFO)
file_handler.setFormatter(CustomFormatter(False))
root_logger.addHandler(file_handler)



#!/usr/bin/env python

from setuptools import setup, find_packages

setup(name='rws',
      version='0.2.0',
      description='Ranking Web Server',
      author='Algorithm Ninja',
      author_email='algorithm@ninja',
      license='AGPL3',
      url='https://github.com/algorithm-ninja/rws',
      packages=find_packages(),
      package_data={
          '': ['*.svg', '*.js', '*.html', '*.css', '*.png', '*.ico']
      },
      install_requires=[
          'six',
          'gevent',
          'werkzeug',
          'cmscommon==0.1.0',
      ],
      entry_points={
          'console_scripts': [
              'rws=cmsranking.RankingWebServer:main'
          ]
      })
